# Copyright (c) 2001, Gemini Data Loggers (UK) Ltd

__version__ = '0.1'

import threading, time, sys, socket, string, atexit, re, binascii, md5, os
import asyncore, asynchat

from Messaging import Messaging
from rflog import log_mixin
import sm, trigger

class _trigger(sm.asyncore_in_server_thread,trigger.trigger):
    pass

_replication_trigger = _trigger()

file_map = {}

_lock = threading.RLock()
_loop_interrupt_lock = threading.RLock()

class _combined_lock:
    def acquire(self):
        _loop_interrupt_lock.acquire()
        _replication_trigger.pull_trigger()
        _lock.acquire()
    def release(self):
        _lock.release()
        _loop_interrupt_lock.release()
_combined_lock = _combined_lock()

try: from posix import fsync
except: fsync=None

generation_count = 0L

class ReplicationError(EnvironmentError):
    pass

class _connection(sm.asyncore_in_server_thread, log_mixin, Messaging):

    # A tiny incoming buffer:
    # Not a problem because we dont handle large volumes of incoming data,
    # however it is better to spin the select loop faster to reduce contention
    # for these big global interlocks.
    ac_in_buffer_size	= 50

    def __init__(self, server, conn, addr):
        self.server = server
        self.addr = addr
        self.type = ''
        self.eager_writes = 10*1024
        Messaging.__init__ (self, conn)
        self.log_info('connected')
        self.push('Ignore hello\n')

    def handle_close(self):
        self.log_info('closing unexpectedly')
        self.close()

    def handle_message_completed(self,cmd,params):
        filename,generation = string.split(params)
        try:
            file = file_map[filename]
        except KeyError:
            # Not a hard error; we may have just moved or closed it
            self.log_info('unexpected file %r in completed' % (filename,))
            return
        else:
            file.completed_by[self._fileno] = generation

    def handle_message_settype(self,cmd,params):
        # This parameter has no well defined meaning. It is
        # provided as an assistance to derived classes with
        # sophisticated implementations of the
        # is_sufficient_replication method
        self.type = params

    def handle_message_crediteagerwrites(self,cmd,params):
        self.eager_writes = max(self.eager_writes+int(params), 10*1024)

    def handle_message_dir(self,cmd,params):
        b = ['Files']
        for name,file in file_map.items():
            pos = file.file.tell()
            file.file.seek(0,2)
            b.append('%s/%s' % (name,file.file.tell()))
            file.file.seek(pos)
        self.push(string.join(b)+'\n')

    def handle_message_read(self,cmd,params):
        filename,start,desired_length,checksum1 = string.split(params)
        start = int(start)
        desired_length = int(desired_length)
        length = self.fix_chunk_size(start,desired_length)
        try:
            file = file_map[filename]
        except KeyError:
            # Not a hard error; we may have just moved or closed it
            self.log_info('unexpected file %r in read' % (filename,))
            return
        else:
            original_pos = file.file.tell()
            file.file.seek(0,2)
            end = file.file.tell()
            if start>=end:
                #self.log_info('Dropping read beyond end of file %s %d %d' % (filename,start,end))
                pass
            else:
                file.file.seek(start)
                data = file.file.read(length)
                file.file.seek(original_pos)
                checksum2 = binascii.b2a_hex(md5.md5(data).digest())
                if checksum1==checksum2 and desired_length==len(data):
                    self.push('Validate %s %s %s\n' % (filename,start,len(data)))
                    #self.log_info('v %s %s' % (start,len(data)))
                else:
                    header = 'Data %s %s %s %s\n' % (filename,start,len(data),checksum2)
                    self.push(header+data+'\n')
                    #self.log_info('d %s %s' % (start,len(data)))

    def fix_chunk_size(self,start,length):
        return min(length,1024*64)

    def log_info(self, message, type='info'):
        message = 'ReplicatedFile connection %r %s' % (self.addr,message)
        log_mixin.log_info(self,message,type)
        

class _server(sm.asyncore_in_server_thread, log_mixin, asyncore.dispatcher):

    def __init__(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind(('',int(os.environ.get('REPLICATED_FILE_PORT','2202'))))
        self.listen(5)

    def writable(self): return 0
    
    def handle_read(self): pass
    
    def readable(self): return 1
    
    def handle_connect (self): pass

    def handle_accept(self):
        try:
            sock, addr = self.accept()
        except socket.error:
            self.log_info('warning: accept failed\n')
        _connection(self,sock,addr)

    def log_info(self, message, type='info'):
        message = 'ReplicatedFile server %s' % (message)
        log_mixin.log_info(self,message,type)

class _server_thread(threading.Thread):
    """Singleton thread to run the Replication Server select loop.

    Thread is started when the module is imported, and terminated
    using atexit.
    """

    def __init__(self):
        threading.Thread.__init__(self)
        self.server = _server()

    def run(self):
        self.server.log_info('starting thread')
        while sm.socket_map:
            _lock.acquire()
            try:
                asyncore.poll(map=sm.socket_map,timeout=30)
            finally:
                _lock.release()
            _loop_interrupt_lock.acquire()
            _loop_interrupt_lock.release()
        self.server.log_info('ending thread')

    def _tidyup(self):
        self.server.log_info('triggering atexit tidyup')
        _replication_trigger.pull_trigger(self._closeall)

    def _closeall(self):
        self.server.log_info('starting atexit tidyup')
        asyncore.close_all(map=sm.socket_map)
        file_map.clear()
        self.server.log_info('finished atexit tidyup')

_server_thread = _server_thread()
_server_thread.start()
atexit.register(_server_thread._tidyup)

# Refuse to deal with files of troublesome names
_simple_file_name = re.compile('^[a-zA-Z0-9][a-zA-Z0-9\.]{0,20}$').match

def _push_to_all(message):
    for x in sm.socket_map.values():
        if isinstance(x,_connection):
            x.push(message)

class File:
    closed = 0
    name = None

    def __init__(self,name,mode):
        self.full_name = os.path.abspath(name)
        if 'b' not in mode or 'a' in mode:
            raise ValueError('bad mode')
        self.file = open(name,mode)
        self.earliest_sync_end_time = 0
        self.completed_by = {}
        self._just_synced = 0
        _combined_lock.acquire()
        try:
            name = string.lower(os.path.split(name)[1])
            if not _simple_file_name(name):
                raise ValueError('ReplicatedFile names contains unexpected characters')
            self.name = name
            if file_map.has_key(name):
                raise ValueError('ReplicatedFile %s already open' % (name,) )
            file_map[name] = self
            self.file.seek(0,2)
            _push_to_all('Opened %s %s\n' % (name,self.file.tell()) )
            self.file.seek(0)
        finally:
            _combined_lock.release()

    def is_sufficient_replication(self,connections):
        """Would the specified connections make up adequate replication?
        """
        # We are happy as long as there is one permanent replica somewhere.
        # Derived classes may overide this.
        for connection in connections:
            if connection.type!='temporary':
                return 1
        return 0

    def time_to_transmit(self,size):
        """How long should we allow to transmit this many bytes to suffient replicas?
        """
        # Assume 100k per second.
        # Derived classes may overide this.
        return size / 102400.0

    def calc_sync_end_time(self):
        """Called at the start of synchronisation. At what time should we give up
        on this as a bad job?
        """
        # Assume that we have accumulated a best guess during writes,
        # but allow at least 2 seconds for round-trip of generation
        # ids. Derived classes may overide this.
        return max(time.time()+2, self.earliest_sync_end_time)

    def close(self):
        _combined_lock.acquire()
        try:
            self._sync()
            self.closed = 1
            self.file.close()
            del file_map[self.name]
            _push_to_all('Closed %s\n' % (self.name,))
        finally:
            _combined_lock.release()

    def write(self,data):
        if not data:
            return
        _combined_lock.acquire()
        try:
            self._just_synced = 0
            pos = self.file.tell()
            self.file.write(data)
            length = len(data)
            eager_cost = length+100
            checksum = None
            self.earliest_sync_end_time = max(self.earliest_sync_end_time,time.time()) + self.time_to_transmit(length)
            for x in sm.socket_map.values():
                if isinstance(x,_connection):
                    if eager_cost<=x.eager_writes:
                        x.eager_writes -= eager_cost
                        if checksum is None:
                            checksum = binascii.b2a_hex(md5.md5(data).digest())
                        header = 'Data %s %s %s %s\n' % (self.name,pos,length,checksum)
                        x.push(header+data+'Echo CreditEagerWrites %s\n' % eager_cost)
                    else:
                        x.push('Invalidate %s %s %s\n' % (self.name, pos, length))
        finally:
            _combined_lock.release()

    def truncate(self,pos=None):
        _combined_lock.acquire()
        try:
            self._just_synced = 0
            if pos is None:
                pos = self.file.tell()
            self.file.truncate(pos)
            _push_to_all('Truncate %s %s\n' % (self.name, pos))
        finally:
            _combined_lock.release()

    def flush(self):
        _combined_lock.acquire()
        try:
            self.file.flush()
            self._sync()
        finally:
            _combined_lock.release()

    def fsync(self):
        _combined_lock.acquire()
        try:
            if fsync is not None:
                fsync(self.file.fileno())
            self._sync()
        finally:
            _combined_lock.release()

    def _sync(self):
        if self._just_synced:
            return
        self._just_synced = 1
        end = self.calc_sync_end_time()
        global generation_count
        generation_count += 1
        generation = str(generation_count)
        _push_to_all('Prioritize %s\nGeneration %s %s\n' % (self.name,generation,self.name))
        try:
            connections = []
            for x in sm.socket_map.values():
                if isinstance(x,_connection):
                    connections.append(x)
            if not self.is_sufficient_replication(connections):
                # And dont bother waiting for new ones to connect
                raise ReplicationError('Insufficient replication clients are available')
            while 1:
                now = time.time()
                if now>end:
                    raise ReplicationError('Timeout waiting for replication clients to sync')
                asyncore.poll(map=sm.socket_map,timeout=min(5,end-now))
                connections = []
                for x in sm.socket_map.values():
                    if isinstance(x,_connection):
                         if self.completed_by.get(x._fileno)==str(generation):
                              connections.append(x)
                if self.is_sufficient_replication(connections):
                    break
        finally:
            _push_to_all('Prioritize\nGeneration %sx\n' % (generation,))

    def rename(self,destination):
        _combined_lock.acquire()
        try:
            if os.path.normpath(os.path.abspath(destination)) == os.path.normpath(os.path.abspath(self.full_name)):
                # Moving a file onto itself?
                return
            old_name = self.name
            name = string.lower(os.path.split(destination)[1])
            if not _simple_file_name(name):
                raise ValueError('ReplicatedFile names contains unexpected characters')
            try:
                overwritten_file = file_map[name]
            except KeyError:
                pass
            else:
                # Not posix sematics; we cannot continue to use overwritten files
                # because replication relies on *names* not inode
                overwritten_file._sync()
                overwritten_file.file.close()
            if sys.platform[:3]=='win':
                # Windows cant rename an open file
                self.file.close()
                # Nor replace an old file during rename
                try:
                    os.unlink(destination)
                except:
                    pass
            os.rename(self.full_name,destination)
            if sys.platform[:3]=='win':
                self.file = open(destination,'r+b')
            if name!=old_name:
                # Only need to inform replicas if the rename changes file name,
                # and not if only the directory changes.
                del file_map[old_name]
                file_map[name] = self
                self.name = name
                _push_to_all('rename %s %s\n' % (old_name, name))
            self.full_name = os.path.abspath(destination)
        finally:
            _combined_lock.release()

    def unlink(self):
        _combined_lock.acquire()
        try:
            # Not posix sematics; we cannot continue to use unlinked files
            # because replication relies on *names* not inode. Maybe it could
            # continue to work without replication?
            self.file.close()
            os.unlink(self.full_name)
            del file_map[self.name]
            _push_to_all('unlink %s\n' % (self.name,))
        finally:
            _combined_lock.release()

    def read(self,*a,**k):
        _combined_lock.acquire()
        try:
            return apply(self.file.read,a,k)
        finally:
            _combined_lock.release()

    def readline(self,*a,**k):
        _combined_lock.acquire()
        try:
            return apply(self.file.readline,a,k)
        finally:
            _combined_lock.release()

    def seek(self,*a,**k):
        _combined_lock.acquire()
        try:
            return apply(self.file.seek,a,k)
        finally:
            _combined_lock.release()

    def tell(self):
        _combined_lock.acquire()
        try:
            return self.file.tell()
        finally:
            _combined_lock.release()

    def fileno(self):
        # Waaaaah. We want to work with the common idiom of fsync(file.fileno())
        # by intercepting the fsync function, so we have to provide a fileno
        # method which is not generally usefull
        return self

