# Copyright (c) 2001, Gemini Data Loggers (UK) Ltd

"""
ReplicationClient.py [options...]

addr
    address off replication server, either a host name
    or host:port. If not specified port is 2202

common options:

--addr=
    Address of replication server.
    This parameter must be specified.

--port=%(port)s
    TCP port of replication server.
    Default is %(port)s

--type=%(type)s
    Misc information about this replica. The
    replication sever may use this information
    in deciding how many replicas it needs for
    sufficient redundancy. The standard
    ReplicatedFile only checks that this value
    is not 'temporary'
    Default is %(type)s

--dir=
    Directory in which replicated files are stored.
    Default is %(dir)s

--daemon=0 or 1
    Whether it should fork into the background
    Default is %(daemon)s. (not windows)

--verbose=%(verbose)s

--tweaking
    Show help on less common options.
"""

tweaking_help="""
ReplicationClient.py tweaking help

--doublecheck=0 or 1
    Whether it should expend bandwidth on internal
    consistency checks. Default is %(doublecheck)s.

--queue=%(queue)s
    Depth of replication request queue. A larger queue
    may help over high-latency links, but may waste
    bandwidth in transferring unnecessary information.
    Default is %(queue)s

--nice_chunk=%(nice_chunk)s
    Preferred size of a replication request, in bytes.
    A smaller value may help over low bandwidth links.
    Default is %(nice_chunk)s

--biggest_chunk=%(biggest_chunk)s
    Maximum size of a replication request, in bytes.
    A smaller value may help over low bandwidth links.
    Default is %(biggest_chunk)s

--cache=0 or 1
    Whether it should validate the content of files
    in the replication directory (1), or ignore them
    and transfer the whole content (0).
    Default is %(cache)s.

"""

__version__ = '0.1'

import time, sys, socket, string, binascii, md5, os, re, copy, getopt, errno
import asyncore, asynchat

from Messaging import Messaging
import rflog

try: from posix import fsync
except: fsync=None

if __name__=="__main__":
    directory = os.path.split(sys.argv[0])[0]
else:                    
    directory = os.path.split(__file__)[0]
directory = os.path.abspath(directory)

# Refuse to deal with files of troublesome names
_simple_file_name = re.compile('^[a-zA-Z0-9][a-zA-Z0-9\.]{0,20}$').match

class ProtocolError(EnvironmentError):
    pass

def main():
    # TODO: command-line configuration
    options = get_options()
    try:
        os.makedirs(options.dir)
    except:
        pass
    if options.daemon:
        try: import posix
        except: pass
        else:
            import zdaemon
            os.chdir(options.dir)
            zdaemon.run(sys.argv, os.path.join(options.dir,'replication-pid'))

    log = rflog.do_pluggable_logging = sized_logger(os.path.join(options.dir,'replication-log'))
    if options.verbose>=1:
        print >> log, 'options are:'
        for name in dir(options):
            print >> log, '%-20s %s' % (name,getattr(options,name))
    try:
        while 1:
            c = replication_client(options)
            while asyncore.socket_map:
                asyncore.poll(30)
                c.tick()
            # that socket has been closed
            if c.try_reconnect:
                time.sleep(3)
            else:
                break
    except KeyboardInterrupt:
        c.log_info('terminating')
        sys.exit(0)
    sys.exit(1)

class options:
    def __init__(self):
        self.addr = ''
        self.port = 2202
        self.queue = 10
        self.nice_chunk = 1024*64
        self.biggest_chunk = self.nice_chunk*3/2
        self.dir = os.path.join(directory,'dest')
        self.doublecheck = 1
        self.verbose = 0
        self.cache = 1
        self.type = 'unspecified'
        self.daemon = 1

def get_options():
    app_options = options()
    if len(sys.argv)<2:
        usage()
        sys.exit(1)
    longopts = [n+'=' for n in dir(app_options) if n[0]!='_']
    longopts.append('tweaking')
    opts,args = getopt.getopt(sys.argv[1:], 'v', longopts )
    if args:
        usage()
        sys.exit(1)
    for o,v in opts:
        if o=='--tweaking':
            sys.stderr.write(tweaking_help % options().__dict__)
            sys.exit(1)
        old = getattr(app_options,o[2:])
        if type(old)==type(0):
            v = int(v)
        setattr(app_options,o[2:],v)
    if app_options.biggest_chunk<app_options.nice_chunk:
        app_options.biggest_chunk = app_options.nice_chunk
    if not app_options.addr:
        usage()
        sys.exit(1)
    return app_options

def usage():
    sys.stderr.write(__doc__ % options().__dict__)

class replication_client(rflog.log_mixin,Messaging):
    def __init__(self,options):
        Messaging.__init__(self)
        self.start_time = time.time()
        self.files = {}
        self.options = options
        self.create_socket(socket.AF_INET,socket.SOCK_STREAM)
        self.connect((options.addr,int(options.port)))
        self.requested = {}
        self.complete = 0
        self.maturity = 0
        self.bytes_updated = self.bytes_validated = 0L
        self.generation = '-'
        self.generation_files = []
        self.try_reconnect = 0

    def tick(self):
        pass

    def handle_connect(self):
        # Announce ourselves
        self.push('SetType %s\n' % (self.options.type,))
        # Send a request for directory information
        self.push('Dir\n')

    def handle_close(self):
        self.log_info('closing unexpectedly')
        self.close()

    def recv(*args):
        self = args[0]
        try:
            return apply(Messaging.recv,args)
        except socket.error,e:
            if e[0] ==errno.ECONNREFUSED:
                return self.connection_refused()
            else:
                raise

    def connection_refused(self):
        self.close()
        self.try_reconnect = 1
        return ''

    def handle_message_opened(self,cmd,params):
        name,size = string.split(params)
        size = int(size)
        self.files[name] = _file(name,size,self.options)
        if size:
            # If the file is non-zero length then we need to
            # fill its content
            self.start_time = time.time()
            self.maturity = min(1,self.maturity)
        self.log_info('opened %s %s' % (name,size))
        if self.options.doublecheck:
            self.push('Dir\n')
        self.consider_queue()

    def handle_message_closed(self,cmd,params):
        del self.files[params]
        self.log_info('closed %s' % (params,))
        if self.options.doublecheck:
            self.push('Dir\n')

    def handle_message_prioritize(self,cmd,params):
        for file in self.files.values():
            file.priority = 0
        for name in string.split(params):
            try:
                self.files[name].priority = 1
            except KeyError:
                if maturity:
                    raise ProtocolError('ERROR: unexpected file %r in prioritize' % (name,))

    def handle_message_files(self,cmd,params):
        self.maturity = max(1,self.maturity)
        for file in self.files.values():
            file.mark = 0
        for entry in string.split(params):
            name,size = string.split(entry,'/')
            size = int(size)
            if self.files.has_key(name):
                file = self.files[name]
                if file.remote_size!=size:
                    raise ProtocolError('ERROR: size mismatch %s %s %s' % (name,size,file.remote_size))
                file.mark = 1
            else:
                file = self.files[name] = _file(name,size,self.options)
                file.mark = 1
        for file in self.files.values():
            if not file.mark:
                raise ProtocolError('ERROR: orphaned file %s' % (file.name))
        self.consider_queue()

    def handle_message_truncate(self,cmd,params):
        name,length = string.split(params)
        length = int(length)
        try:
            file = self.files[name]
        except KeyError:
            if self.maturity:
                raise ProtocolError('ERROR: unexpected file %r in truncate' % (name,))
        else:
            file.file.seek(length)
            file.file.truncate()
            if length>file.remote_size:
                file.invalid.add(file.remote_size,length)
            if length<file.remote_size:
                file.invalid.sub(length,file.remote_size)
            self.check_complete(file)
            file.remote_size = length
            self.log_info('truncated %s to %s' % (name,length))
            self.consider_queue()

    def handle_message_rename(self,cmd,params):
        old,new = string.split(params)
        try:
            file = self.files[old]
        except KeyError:
            if self.maturity:
                raise ProtocolError('ERROR: unexpected file %r in rename' % (old,))
        else:
            try:
                replaced = self.files[new]
            except KeyError:
                self.log_info('renaming %s to new file %s' % (old,new) )
            else:
                self.log_info('renaming %s to replace %s' % (old,new) )
                replaced.file.close()
                del self.files[new]
            file.rename(new,self.options)
            del self.files[old]
            self.files[new] = file

    def handle_message_unlink(self,cmd,params):
        name = params
        try:
            file = self.files[name]
        except KeyError:
            if self.maturity:
                raise ProtocolError('ERROR: unexpected file %r in unlink' % (name,))
        else:
            path = file.path
            del self.files[name]
            file.file.close()
            # Do we unlink this file, to avoid turds in the replica,
            # or leave the turd around in case it can be used to speed up
            # a future replica under this file name?
            os.unlink(path)
            self.log_info('unlinking %s' % (name,))

    def handle_message_invalidate(self,cmd,params):
        name,pos,length = string.split(params)
        pos = int(pos)
        length = int(length)
        try:
            file = self.files[name]
        except KeyError:
            if self.maturity:
                raise ProtocolError('ERROR: unexpected file %r in invalidate' % (name,))
        else:
            file.remote_size = max(file.remote_size,pos+length)
            file.invalid.add(pos,pos+length)
            if self.options.verbose>=2:
                self.log_info('invalidated %s %s %s' % (name,pos,length))
            self.consider_queue()

    def handle_message_generation(self,cmd,params):
        params = string.split(params)
        self.generation = params[0]
        self.generation_files = params[1:]
        for file in self.files.values():
            self.check_complete(file)

    def handle_message_validate(self,cmd,params):
        name,pos,length = string.split(params)
        pos = int(pos)
        length = int(length)
        try:
            file = self.files[name]
        except KeyError:
            if self.maturity:
                raise ProtocolError('Error unexpected file %r in Validate' % (name,))
        else:
            if self.options.verbose>=2:
                self.log_info('validated %s %s %s' % (name,pos,length))
            self.bytes_validated += length
            file.remote_size = max(file.remote_size,pos+length)
            if length:
                file.invalid.sub(pos,pos+length)
            self.check_complete(file)

    def handle_message_data(self,cmd,params,data=None):
        name,pos,length,checksum1 = string.split(params)
        pos = int(pos)
        length = int(length)
        if data is None:
            self.read_data(cmd,params,length)
        else:
            checksum2 = binascii.b2a_hex(md5.md5(data).digest())
            if length!=len(data):
                raise ProtocolError('ERROR: length wrong %s %s' % (length,len(data)))
            if checksum1!=checksum2:
                raise ProtocolError('ERROR: transmission checksum failed')
            try:
                file = self.files[name]
            except KeyError:
                if self.maturity:
                    raise ProtocolError('ERROR: unexpected file %r in Data' % (name,))
            else:
                file.file.seek(pos)
                file.file.write(data)
                if self.options.verbose>=2:
                    self.log_info('updated %s %s %s' % (name,pos,length))
                self.bytes_updated += length
                file.remote_size = max(file.remote_size,pos+length)
                file.invalid.sub(pos,pos+length)
                self.check_complete(file)

    def check_complete(self,file):
        if file.name in self.generation_files and not file.invalid:
            self.push('Completed %s %s\n' % (file.name,self.generation))
            self.generation_files.remove(file.name)
            file.file.flush()
            if fsync: fsync(file.file.fileno())
            if self.options.verbose>=1:
                self.log_info('sync of %s in %s' % (file.name,self.generation) )

    def handle_message_creditqueue(self,cmd,params):
        try:
            del self.requested[params]
        except KeyError:
            raise ProtocolError('ERROR: bad queue credit')
        self.consider_queue()

    def consider_queue(self):
        if not self.maturity:
            return
        while len(self.requested)<self.options.queue:
            # First, find all incomplete files
            files = [file for file in self.files.values() if file.invalid]
            if not files:
                # Nothing to do
                if not self.complete:
                    if self.options.verbose>=1 or self.maturity<2:
                        self.log_info('complete (%s bytes updated. %s validated.)' % (self.bytes_updated,self.bytes_validated))
                    if self.maturity<2:
                        self.push('log synchronisation complete (%d%% updated)\n' %(100*self.bytes_updated/(self.bytes_updated+self.bytes_validated+1)),)
                        self.log_info('elapsed time %s' % self.time_since_start())
                    self.bytes_updated = self.bytes_validated = 0L
                self.maturity = 2
                self.complete = 1
                return
            else:
                if self.complete:
                    if self.options.verbose>=1:
                        self.log_info('incomplete')
                    self.complete = 0
                # Next, find the most incomplete file to request something for
                files_and_sizes = []
                for file in files:
                    # Measure how much remains invalid
                    size = file.invalid.size()
                    # Subtract the size of any requests for this file
                    for name,start,end in self.requested.values():
                        if name==file.name:
                            size -= end-start
                    # Only proceed if some of the file remains unrequested
                    if size>0:
                        files_and_sizes.append((size,file.priority,file))
                if not files_and_sizes:
                    # Nothing to do at this moment.
                    return
                else:
                    # Some files have incomplete, unrequested bytes.
                    # Find the one with the most, and request that.
                    files_and_sizes.sort()
                    file = files_and_sizes[-1][-1]
                    # Work out what remains unrequested for this file
                    invalid = file.invalid.copy()
                    for name,start,end in self.requested.values():
                        if name==file.name:
                            invalid.sub(start,end)
                    assert invalid, 'Some of the file must remain unrequested'
                    # Request the next chunk for this file
                    start,end = invalid.choose_chunk(self.options)
                    file.file.seek(start)
                    if self.options.cache:
                        data = file.file.read(end-start)
                        checksum = binascii.b2a_hex(md5.md5(data).digest())
                    else:
                        checksum = '-'
                    triplet = file.name,start,end
                    request_id = str(id(triplet))
                    self.push('Read %s %s %s %s\nEcho CreditQueue %s\n' % (file.name,start,end-start,checksum,request_id))
                    self.requested[request_id] = triplet

    def time_since_start(self):
        elapsed = time.time()-self.start_time
        if elapsed<240:
            return '%d seconds' % elapsed
        else:
            return '%d minutes' % (elapsed/60.0)
            
    def log_info(self, message, type='info'):
        message = 'ReplicationClient %s' % (message)
        rflog.log_mixin.log_info(self,message,type)


class _file:

    def __init__(self,name,size,options):
        self.name = name
        self.remote_size = size
        self.options = options
        self.priority = 0
        self.path = os.path.abspath(os.path.join(options.dir,name))
        try:
            self.file = open(self.path,'r+b')
        except IOError:
            self.file = open(self.path,'w+b')
        self.file.truncate(size)
        self.invalid = _region()
        if size:
            self.invalid.add(0,size)

    def rename(self,name,options):
        new_path = os.path.abspath(os.path.join(options.dir,name))
        if sys.platform[:3]=='win':
            # Windows cant replace an old file during rename
            try:
                os.unlink(new_path)
            except:
                pass
            # Nor rename an open file
            self.file.close()
        os.rename(self.path,new_path)
        self.name = name
        self.path = new_path
        if sys.platform[:3]=='win':
            self.file = open(self.path,'r+b')

class _region:
    """A class to keep track of which regions of a file contain
    validated bytes, and which regions contain little more than
    hints to use for saving bandwidth during validation.
    """

    def __init__(self):
        # self._i is a list of (start,end) pairs which designate the range of
        # file segments that make up this region. segments do not overlap,
        # are never empty, and are in no particular order.
        self._i = ()

    def copy(self):
        return copy.copy(self)

    def add(self,start,end):
        """Add a band to this region"""
        i = []
        assert start<end, 'non-empty region'
        for s,e in self._i:
            if e<start or end<s:
                i.append((s,e))
            else:
                start = min(s,start)
                end = max(e,end)
        assert start<end, 'non-empty region'
        i.append((start,end))
        self._i = tuple(i)

    def sub(self,start,end):
        """Remove a band from this region"""
        i = []
        assert start<end, 'non-empty region'
        for s,e in self._i:
            if e<start or end<s:
                i.append((s,e))
            else:
                if s<start:
                    p = min(e,start)
                    assert s<p, 'non-empty region'
                    i.append((s,p))
                if e>end:
                    p = max(s,end)
                    assert p<e, 'non-empty region'
                    i.append((p,e))
        self._i = tuple(i)

    def __nonzero__(self):
        return len(self._i)

    def size(self):
        total = 0
        for s,e in self._i:
            assert s<e, 'non-empty region'
            total = total+e-s
        return total

    def choose_chunk(self,options):
        if not self._i:
            raise ValueError('empty region; no chunks')
        # Return a sensibly-sized chunk from this region.
        # Work on the smallest segment first, to reduce segment fragmentation
        smallest_start = smallest_end = 0
        smallest_size = 'this value is bigger than any number'
        assert 1000000<smallest_size, 'all numbers compare less than strings'
        for start,end in self._i:
            size = end-start
            if size<smallest_size:
                smallest_start,smallest_end,smallest_size = start,end,size
                if size<=options.biggest_chunk:
                    break
        if end-start>options.biggest_chunk:
            end = start+options.nice_chunk
        assert start<end, 'non-empty region'
        return start,end

class sized_logger:
    def __init__ (self,file,maxsize=500000,mode='a'):
        self.filename = file
        self.mode = mode
        self.file = open(file,mode)
        self.maxsize = maxsize

    def __call__(self,message,type):
        timestamp = time.strftime('%Y-%m-%dT%H:%M:%S', time.gmtime(time.time()) )
        self.write ('%s %s - %s\n' % (timestamp, type, message))

    def write(self,message):
        self.file.write(message)
        self.file.flush()
        self.maybe_reset()

    def maybe_reset(self):
        try:
            size = self.file.tell()
            do_reset = size>self.maxsize
        except os.error:
            do_reset=1
        if do_reset:
            self.reset()

    def reset(self):
        old_filename = self.filename+'.old'
        try:
            os.unlink(old_filename)
        except os.error:
            pass
        self.file.close()
        os.rename(self.filename,old_filename)
        self.file = open(self.filename,self.mode)

    def flush (self):
        self.file.flush()

if __name__=='__main__':
    main()